@@ -62,7 +62,9 @@ typedef enum {
 typedef struct {
         SoupClientMessageIO iface;
 
-        SoupConnection *conn;
+        GThread *owner;
+        gboolean async;
+        GWeakRef conn;
         GIOStream *stream;
         GInputStream *istream;
         GOutputStream *ostream;
@@ -116,7 +118,6 @@ typedef struct {
         gpointer completion_data;
         SoupHTTP2IOState state;
         GError *error;
-        uint32_t http2_error;
         gboolean paused;
         guint32 stream_id;
         gboolean can_be_restarted;
@@ -268,20 +269,6 @@ set_error_for_data (SoupHTTP2MessageData *data,
                 g_error_free (error);
 }
 
-static void
-set_http2_error_for_data (SoupHTTP2MessageData *data,
-                          uint32_t              error_code)
-{
-        h2_debug (data->io, data, "[SESSION] Error: %s", nghttp2_http2_strerror (error_code));
-
-        if (data->error)
-                return;
-
-        data->http2_error = error_code;
-        data->error = g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED,
-                                   "HTTP/2 Error: %s", nghttp2_http2_strerror (error_code));
-}
-
 static void
 set_io_error (SoupClientMessageIOHTTP2 *io,
               GError                   *error)
@@ -292,11 +279,6 @@ set_io_error (SoupClientMessageIOHTTP2 *io,
                 io->error = error;
         else
                 g_error_free (error);
-
-        if (io->close_task && !io->goaway_sent) {
-                g_task_return_boolean (io->close_task, TRUE);
-                g_clear_object (&io->close_task);
-        }
 }
 
 static void
@@ -323,22 +305,6 @@ advance_state_from (SoupHTTP2MessageData *data,
         data->state = to;
 }
 
-static gboolean
-soup_http2_message_data_can_be_restarted (SoupHTTP2MessageData *data,
-                                          GError               *error)
-{
-        if (data->can_be_restarted)
-                return TRUE;
-
-        return data->state < STATE_READ_DATA_START &&
-                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT) &&
-                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK) &&
-                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED) &&
-                error->domain != G_TLS_ERROR &&
-                data->http2_error == NGHTTP2_NO_ERROR &&
-                SOUP_METHOD_IS_IDEMPOTENT (soup_message_get_method (data->msg));
-}
-
 static void
 soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
 {
@@ -350,7 +316,6 @@ soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
         if (g_cancellable_set_error_if_cancelled (g_task_get_cancellable (task), &error)) {
                 io->pending_io_messages = g_list_remove (io->pending_io_messages, data);
                 data->task = NULL;
-                soup_client_message_io_http2_finished ((SoupClientMessageIO *)io, msg);
                 g_task_return_error (task, error);
                 g_object_unref (task);
                 return;
@@ -365,7 +330,7 @@ soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
         if (data->error) {
                 GError *error = g_steal_pointer (&data->error);
 
-                if (soup_http2_message_data_can_be_restarted (data, error))
+                if (data->can_be_restarted)
                         data->item->state = SOUP_MESSAGE_RESTARTING;
                 else
                         soup_message_set_metrics_timestamp (data->msg, SOUP_MESSAGE_METRICS_RESPONSE_END);
@@ -434,7 +399,7 @@ io_write_ready (GObject                  *stream,
                 return G_SOURCE_REMOVE;
         }
 
-        while (!error && nghttp2_session_want_write (io->session))
+        while (nghttp2_session_want_write (io->session) && !error)
                 io_write (io, FALSE, NULL, &error);
 
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
@@ -462,16 +427,14 @@ io_try_write (SoupClientMessageIOHTTP2 *io,
                 if (blocking || !nghttp2_session_want_write (io->session))
                         return;
         } else {
-                while (!error && nghttp2_session_want_write (io->session))
+                while (nghttp2_session_want_write (io->session) && !error)
                         io_write (io, blocking, NULL, &error);
         }
 
-        if (io->in_callback || g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
+        if (!blocking && (io->in_callback || g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK))) {
                 g_clear_error (&error);
                 io->write_source = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (io->ostream), NULL);
                 g_source_set_name (io->write_source, "Soup HTTP/2 write source");
-                /* Give write more priority than read */
-                g_source_set_priority (io->write_source, G_PRIORITY_DEFAULT - 1);
                 g_source_set_callback (io->write_source, (GSourceFunc)io_write_ready, io, NULL);
                 g_source_attach (io->write_source, g_main_context_get_thread_default ());
         }
@@ -490,24 +453,14 @@ io_read (SoupClientMessageIOHTTP2  *io,
         gssize read;
         int ret;
 
-        /* Always try to write before read, in case there's a pending reset stream after an error. */
-        io_try_write (io, blocking);
-
         if ((read = g_pollable_stream_read (io->istream, buffer, sizeof (buffer),
                                             blocking, cancellable, error)) < 0)
             return FALSE;
 
-        if (read == 0) {
-                g_set_error_literal (error, G_IO_ERROR,
-                                     G_IO_ERROR_PARTIAL_INPUT,
-                                     _("Connection terminated unexpectedly"));
-                return FALSE;
-        }
-
         g_warn_if_fail (io->in_callback == 0);
         ret = nghttp2_session_mem_recv (io->session, buffer, read);
         NGCHECK (ret);
-        return ret > 0;
+        return ret != 0;
 }
 
 static gboolean
@@ -516,6 +469,7 @@ io_read_ready (GObject                  *stream,
 {
         GError *error = NULL;
         gboolean progress = TRUE;
+        SoupConnection *conn;
 
         if (io->error) {
                 g_clear_pointer (&io->read_source, g_source_unref);
@@ -525,10 +479,11 @@ io_read_ready (GObject                  *stream,
         /* Mark the connection as in use to make sure it's not disconnected while
          * processing pending messages, for example if a goaway is received.
          */
-        if (io->conn)
-                soup_connection_set_in_use (io->conn, TRUE);
+        conn = g_weak_ref_get (&io->conn);
+        if (conn)
+                soup_connection_set_in_use (conn, TRUE);
 
-        while (progress && nghttp2_session_want_read (io->session)) {
+        while (nghttp2_session_want_read (io->session) && progress) {
                 progress = io_read (io, FALSE, NULL, &error);
                 if (progress) {
                         g_list_foreach (io->pending_io_messages,
@@ -539,13 +494,13 @@ io_read_ready (GObject                  *stream,
 
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
                 g_error_free (error);
-                if (io->conn)
-                        soup_connection_set_in_use (io->conn, FALSE);
+                if (conn) {
+                        soup_connection_set_in_use (conn, FALSE);
+                        g_object_unref (conn);
+                }
                 return G_SOURCE_CONTINUE;
         }
 
-        io->is_shutdown = TRUE;
-
         if (error) {
                 set_io_error (io, error);
                 g_list_foreach (io->pending_io_messages,
@@ -553,9 +508,13 @@ io_read_ready (GObject                  *stream,
                                 NULL);
         }
 
+        io->is_shutdown = TRUE;
+
         g_clear_pointer (&io->read_source, g_source_unref);
-        if (io->conn)
-                soup_connection_set_in_use (io->conn, FALSE);
+        if (conn) {
+                soup_connection_set_in_use (conn, FALSE);
+                g_object_unref (conn);
+        }
         return G_SOURCE_REMOVE;
 }
 
@@ -595,7 +554,7 @@ soup_client_message_io_http2_terminate_session (SoupClientMessageIOHTTP2 *io)
 
         io->session_terminated = TRUE;
         NGCHECK (nghttp2_session_terminate_session (io->session, NGHTTP2_NO_ERROR));
-        io_try_write (io, FALSE);
+        io_try_write (io, !io->async);
 }
 
 /* HTTP2 read callbacks */
@@ -712,7 +671,8 @@ handle_goaway (SoupClientMessageIOHTTP2 *io,
                 if ((error_code == 0 && (int32_t)data->stream_id > last_stream_id) ||
                      data->state < STATE_READ_DONE) {
                         /* TODO: We can restart unfinished messages */
-                        set_http2_error_for_data (data, error_code);
+                        set_error_for_data (data, g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED,
+                                            "HTTP/2 Error: %s", nghttp2_http2_strerror (error_code)));
                 }
         }
 }
@@ -728,7 +688,7 @@ on_frame_recv_callback (nghttp2_session     *session,
         io->in_callback++;
 
         if (frame->hd.stream_id == 0) {
-                h2_debug (io, NULL, "[RECV] [%s] Recieved (%u)", frame_type_to_string (frame->hd.type), frame->hd.flags);
+                h2_debug (io, NULL, "[RECV] [%s] Received (%u)", frame_type_to_string (frame->hd.type), frame->hd.flags);
 
                 switch (frame->hd.type) {
                 case NGHTTP2_GOAWAY:
@@ -751,7 +711,7 @@ on_frame_recv_callback (nghttp2_session     *session,
         }
 
         data = nghttp2_session_get_stream_user_data (session, frame->hd.stream_id);
-        h2_debug (io, data, "[RECV] [%s] Recieved (%u)", frame_type_to_string (frame->hd.type), frame->hd.flags);
+        h2_debug (io, data, "[RECV] [%s] Received (%u)", frame_type_to_string (frame->hd.type), frame->hd.flags);
 
         if (!data) {
                 if (!(frame->hd.flags & NGHTTP2_FLAG_END_STREAM) && frame->hd.type != NGHTTP2_RST_STREAM)
@@ -826,8 +786,10 @@ on_frame_recv_callback (nghttp2_session     *session,
                 }
                 break;
         case NGHTTP2_RST_STREAM:
-                if (frame->rst_stream.error_code != NGHTTP2_NO_ERROR)
-                        set_http2_error_for_data (data, frame->rst_stream.error_code);
+                if (frame->rst_stream.error_code != NGHTTP2_NO_ERROR) {
+                        set_error_for_data (data, g_error_new_literal (G_IO_ERROR, G_IO_ERROR_FAILED,
+                                                                       nghttp2_http2_strerror (frame->rst_stream.error_code)));
+                }
                 break;
         case NGHTTP2_WINDOW_UPDATE:
                 h2_debug (io, data, "[RECV] WINDOW_UPDATE: increment=%d, total=%d", frame->window_update.window_size_increment,
@@ -857,7 +819,7 @@ on_data_chunk_recv_callback (nghttp2_session *session,
 
         io->in_callback++;
 
-        h2_debug (io, msgdata, "[DATA] Recieved chunk, len=%zu, flags=%u, paused=%d", len, flags, msgdata->paused);
+        h2_debug (io, msgdata, "[DATA] Received chunk, len=%zu, flags=%u, paused=%d", len, flags, msgdata->paused);
 
         g_assert (msgdata->body_istream != NULL);
         soup_body_input_stream_http2_add_data (SOUP_BODY_INPUT_STREAM_HTTP2 (msgdata->body_istream), data, len);
@@ -957,7 +919,15 @@ on_frame_send_callback (nghttp2_session     *session,
                 break;
         case NGHTTP2_RST_STREAM:
                 h2_debug (io, data, "[SEND] [RST_STREAM] stream_id=%u", frame->hd.stream_id);
-                g_hash_table_foreach_remove (io->closed_messages, (GHRFunc)remove_closed_stream, (gpointer)frame);
+                if (g_hash_table_foreach_remove (io->closed_messages, (GHRFunc)remove_closed_stream, (gpointer)frame)) {
+                        SoupConnection *conn = g_weak_ref_get (&io->conn);
+
+                        if (conn) {
+                                soup_connection_set_in_use (conn, FALSE);
+                                g_object_unref (conn);
+                        }
+                }
+
                 break;
         case NGHTTP2_GOAWAY:
                 h2_debug (io, data, "[SEND] [%s]", frame_type_to_string (frame->hd.type));
@@ -969,7 +939,7 @@ on_frame_send_callback (nghttp2_session     *session,
                         source = g_idle_source_new ();
                         g_source_set_name (source, "Soup HTTP/2 close source");
                         g_source_set_callback (source, (GSourceFunc)close_in_idle_cb, io, NULL);
-                        g_source_attach (source, g_main_context_get_thread_default ());
+                        g_source_attach (source, g_task_get_context (io->close_task));
                         g_source_unref (source);
                 }
                 break;
@@ -982,17 +952,45 @@ on_frame_send_callback (nghttp2_session     *session,
         return 0;
 }
 
+static gboolean
+update_connection_in_use (gpointer        key,
+                          gpointer        value,
+                          SoupConnection *conn)
+{
+        soup_connection_set_in_use (conn, FALSE);
+
+        return TRUE;
+}
+
+static void
+process_pending_closed_messages (SoupClientMessageIOHTTP2 *io)
+{
+        SoupConnection *conn = g_weak_ref_get (&io->conn);
+
+        if (!conn) {
+                g_hash_table_remove_all (io->closed_messages);
+                return;
+        }
+
+        g_hash_table_foreach_remove (io->closed_messages, (GHRFunc)update_connection_in_use, conn);
+        g_object_unref (conn);
+}
+
 static int
 on_frame_not_send_callback (nghttp2_session     *session,
                             const nghttp2_frame *frame,
                             int                  lib_error_code,
                             void                *user_data)
 {
+        SoupClientMessageIOHTTP2 *io = user_data;
         SoupHTTP2MessageData *data = nghttp2_session_get_stream_user_data (session, frame->hd.stream_id);
 
-        h2_debug (user_data, data, "[SEND] [%s] Failed: %s", frame_type_to_string (frame->hd.type),
+        h2_debug (io, data, "[SEND] [%s] Failed: %s", frame_type_to_string (frame->hd.type),
                   nghttp2_strerror (lib_error_code));
 
+        if (lib_error_code == NGHTTP2_ERR_SESSION_CLOSING)
+                process_pending_closed_messages (io);
+
         return 0;
 }
 
@@ -1011,19 +1009,14 @@ on_stream_close_callback (nghttp2_session *session,
         data->io->in_callback++;
 
         switch (error_code) {
-        case NGHTTP2_NO_ERROR:
-                break;
         case NGHTTP2_REFUSED_STREAM:
-                if (data->state < STATE_READ_DATA_START)
+                if (data->state < STATE_READ_DATA)
                         data->can_be_restarted = TRUE;
                 break;
         case NGHTTP2_HTTP_1_1_REQUIRED:
                 soup_message_set_force_http_version (data->item->msg, SOUP_HTTP_1_1);
                 data->can_be_restarted = TRUE;
                 break;
-        default:
-                set_http2_error_for_data (data, error_code);
-                break;
         }
 
         data->io->in_callback--;
@@ -1459,7 +1452,6 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 	SoupMessageIOCompletionFn completion_cb;
 	gpointer completion_data;
         SoupMessageIOCompletion completion;
-        gboolean is_closed;
 
         data = get_data_for_message (io, msg);
 
@@ -1472,10 +1464,11 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 
 	g_object_ref (msg);
 
-        is_closed = nghttp2_session_get_stream_user_data (io->session, data->stream_id) == NULL;
         nghttp2_session_set_stream_user_data (io->session, data->stream_id, NULL);
 
-        if (!io->is_shutdown && !is_closed) {
+        if (!io->is_shutdown) {
+                SoupConnection *conn;
+
                 NGCHECK (nghttp2_submit_rst_stream (io->session, NGHTTP2_FLAG_NONE, data->stream_id,
                                                     completion == SOUP_MESSAGE_IO_COMPLETE ? NGHTTP2_NO_ERROR : NGHTTP2_CANCEL));
                 soup_http2_message_data_close (data);
@@ -1484,6 +1477,14 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
                         g_warn_if_reached ();
                 if (!g_hash_table_add (io->closed_messages, data))
                         g_warn_if_reached ();
+
+                conn = g_weak_ref_get (&io->conn);
+                if (conn) {
+                        soup_connection_set_in_use (conn, TRUE);
+                        g_object_unref (conn);
+                }
+
+                io_try_write (io, !io->async);
         } else {
                 if (!g_hash_table_remove (io->messages, msg))
                         g_warn_if_reached ();
@@ -1494,12 +1495,8 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 
 	g_object_unref (msg);
 
-        if (io->is_shutdown) {
+        if (io->is_shutdown)
                 soup_client_message_io_http2_terminate_session (io);
-                return;
-        }
-
-        io_try_write (io, FALSE);
 }
 
 static void
@@ -1634,12 +1631,13 @@ io_run (SoupHTTP2MessageData *data,
         GCancellable         *cancellable,
         GError              **error)
 {
+        SoupClientMessageIOHTTP2 *io = data->io;
         gboolean progress = FALSE;
 
-        if (data->state < STATE_WRITE_DONE && nghttp2_session_want_write (data->io->session))
-                progress = io_write (data->io, TRUE, cancellable, error);
-        else if (data->state < STATE_READ_DONE && nghttp2_session_want_read (data->io->session))
-                progress = io_read (data->io, TRUE, cancellable, error);
+        if (data->state < STATE_WRITE_DONE && !io->in_callback && nghttp2_session_want_write (io->session))
+                progress = io_write (io, TRUE, cancellable, error);
+        else if (data->state < STATE_READ_DONE && !io->in_callback && nghttp2_session_want_read (io->session))
+                progress = io_read (io, TRUE, cancellable, error);
 
         return progress;
 }
@@ -1670,13 +1668,11 @@ io_run_until (SoupClientMessageIOHTTP2 *io,
                 progress = io_run (data, cancellable, &my_error);
 
         if (my_error) {
-                io->is_shutdown = TRUE;
-                set_io_error (io, my_error);
+                g_propagate_error (error, my_error);
+                g_object_unref (msg);
+                return FALSE;
         }
 
-        if (io->error && !data->error)
-                data->error = g_error_copy (io->error);
-
 	if (data->error) {
                 g_propagate_error (error, g_steal_pointer (&data->error));
 		g_object_unref (msg);
@@ -1705,13 +1701,12 @@ soup_client_message_io_http2_run_until_read (SoupClientMessageIO  *iface,
 {
         SoupClientMessageIOHTTP2 *io = (SoupClientMessageIOHTTP2 *)iface;
         SoupHTTP2MessageData *data = get_data_for_message (io, msg);
-        GError *my_error = NULL;
 
-        if (io_run_until (io, msg, STATE_READ_DATA, cancellable, &my_error))
+        if (io_run_until (io, msg, STATE_READ_DATA, cancellable, error))
                 return TRUE;
 
         if (get_io_data (msg) == io) {
-                if (soup_http2_message_data_can_be_restarted (data, my_error))
+                if (data->can_be_restarted)
                         data->item->state = SOUP_MESSAGE_RESTARTING;
                 else
                         soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_RESPONSE_END);
@@ -1719,8 +1714,6 @@ soup_client_message_io_http2_run_until_read (SoupClientMessageIO  *iface,
                 soup_client_message_io_http2_finished (iface, msg);
         }
 
-        g_propagate_error (error, my_error);
-
         return FALSE;
 }
 
@@ -1773,6 +1766,31 @@ soup_client_message_io_http2_run_until_read_async (SoupClientMessageIO *iface,
                 soup_http2_message_data_check_status (data);
 }
 
+static void
+soup_client_message_io_http2_set_owner (SoupClientMessageIOHTTP2 *io,
+                                        GThread                  *owner)
+{
+        if (owner == io->owner)
+                return;
+
+        io->owner = owner;
+        g_assert (!io->write_source);
+        if (io->read_source) {
+                g_source_destroy (io->read_source);
+                g_source_unref (io->read_source);
+                io->read_source = NULL;
+        }
+
+        io->async = g_main_context_is_owner (g_main_context_get_thread_default ());
+        if (!io->async)
+                return;
+
+        io->read_source = g_pollable_input_stream_create_source (G_POLLABLE_INPUT_STREAM (io->istream), NULL);
+        g_source_set_name (io->read_source, "Soup HTTP/2 read source");
+        g_source_set_callback (io->read_source, (GSourceFunc)io_read_ready, io, NULL);
+        g_source_attach (io->read_source, g_main_context_get_thread_default ());
+}
+
 static gboolean
 soup_client_message_io_http2_close_async (SoupClientMessageIO *iface,
                                           SoupConnection      *conn,
@@ -1783,9 +1801,18 @@ soup_client_message_io_http2_close_async (SoupClientMessageIO *iface,
         if (io->goaway_sent)
                 return FALSE;
 
-        g_assert (!io->close_task);
-        io->close_task = g_task_new (conn, NULL, callback, NULL);
+        soup_client_message_io_http2_set_owner (io, g_thread_self ());
+        if (io->async) {
+                g_assert (!io->close_task);
+                io->close_task = g_task_new (conn, NULL, callback, NULL);
+        }
+
         soup_client_message_io_http2_terminate_session (io);
+        if (!io->async) {
+                g_assert (io->goaway_sent);
+                return FALSE;
+        }
+
         return TRUE;
 }
 
@@ -1803,8 +1830,7 @@ soup_client_message_io_http2_destroy (SoupClientMessageIO *iface)
                 g_source_unref (io->write_source);
         }
 
-        if (io->conn)
-                g_object_remove_weak_pointer (G_OBJECT (io->conn), (gpointer*)&io->conn);
+        g_weak_ref_clear (&io->conn);
         g_clear_object (&io->stream);
         g_clear_object (&io->close_task);
         g_clear_pointer (&io->session, nghttp2_session_del);
@@ -1816,6 +1842,14 @@ soup_client_message_io_http2_destroy (SoupClientMessageIO *iface)
         g_free (io);
 }
 
+static void
+soup_client_message_io_http2_owner_changed (SoupClientMessageIO *iface)
+{
+        SoupClientMessageIOHTTP2 *io = (SoupClientMessageIOHTTP2 *)iface;
+
+        soup_client_message_io_http2_set_owner (io, g_thread_self ());
+}
+
 static const SoupClientMessageIOFuncs io_funcs = {
         soup_client_message_io_http2_destroy,
         soup_client_message_io_http2_finished,
@@ -1833,7 +1867,8 @@ static const SoupClientMessageIOFuncs io_funcs = {
         soup_client_message_io_http2_is_open,
         soup_client_message_io_http2_in_progress,
         soup_client_message_io_http2_is_reusable,
-        soup_client_message_io_http2_get_cancellable
+        soup_client_message_io_http2_get_cancellable,
+        soup_client_message_io_http2_owner_changed
 };
 
 G_GNUC_PRINTF(1, 0)
@@ -1893,19 +1928,14 @@ soup_client_message_io_http2_new (SoupConnection *conn)
         SoupClientMessageIOHTTP2 *io = g_new0 (SoupClientMessageIOHTTP2, 1);
         soup_client_message_io_http2_init (io);
 
-        io->conn = conn;
-        g_object_add_weak_pointer (G_OBJECT (io->conn), (gpointer*)&io->conn);
+        g_weak_ref_init (&io->conn, conn);
 
         io->stream = g_object_ref (soup_connection_get_iostream (conn));
         io->istream = g_io_stream_get_input_stream (io->stream);
         io->ostream = g_io_stream_get_output_stream (io->stream);
         io->connection_id = soup_connection_get_id (conn);
 
-        io->read_source = g_pollable_input_stream_create_source (G_POLLABLE_INPUT_STREAM (io->istream), NULL);
-        g_source_set_name (io->read_source, "Soup HTTP/2 read source");
-        g_source_set_priority (io->read_source, G_PRIORITY_DEFAULT);
-        g_source_set_callback (io->read_source, (GSourceFunc)io_read_ready, io, NULL);
-        g_source_attach (io->read_source, g_main_context_get_thread_default ());
+        soup_client_message_io_http2_set_owner (io, soup_connection_get_owner (conn));
 
         NGCHECK (nghttp2_session_set_local_window_size (io->session, NGHTTP2_FLAG_NONE, 0, INITIAL_WINDOW_SIZE));
 
@@ -1915,7 +1945,7 @@ soup_client_message_io_http2_new (SoupConnection *conn)
                 { NGHTTP2_SETTINGS_ENABLE_PUSH, 0 },
         };
         NGCHECK (nghttp2_submit_settings (io->session, NGHTTP2_FLAG_NONE, settings, G_N_ELEMENTS (settings)));
-        io_try_write (io, FALSE);
+        io_try_write (io, !io->async);
 
         return (SoupClientMessageIO *)io;
 }
