@@ -19,29 +19,22 @@
 #include "soup-uri-utils-private.h"
 
 /**
- * SECTION:soup-server-message
- * @short_description: An HTTP server request and response.
- * @see_also: #SoupMessageHeaders, #SoupMessageBody
+ * SoupServerMessage:
+ *
+ * An HTTP server request and response pair.
  *
  * A SoupServerMessage represents an HTTP message that is being sent or
- * received on a #SoupServer
+ * received on a [class@Server].
  *
- * #SoupServer will create #SoupServerMessage<!-- -->s automatically for
+ * [class@Server] will create `SoupServerMessage`s automatically for
  * incoming requests, which your application will receive via handlers.
  *
  * Note that libsoup's terminology here does not quite match the HTTP
- * specification: in RFC 2616, an "HTTP-message" is
- * <emphasis>either</emphasis> a Request, <emphasis>or</emphasis> a
- * Response. In libsoup, a #SoupServerMessage combines both the request and
- * the response.
+ * specification: in RFC 2616, an "HTTP-message" is *either* a Request, *or* a
+ * Response. In libsoup, a #SoupServerMessage combines both the request and the
+ * response.
  **/
 
-/**
- * SoupServerMessage:
- *
- * Class represnting an HTTP request and response pair for a server.
- */
-
 struct _SoupServerMessage {
         GObject             parent;
 
@@ -72,13 +65,16 @@ struct _SoupServerMessage {
         SoupServerMessageIOData *io_data;
 
         gboolean                 options_ping;
+
+        GTlsCertificate      *tls_peer_certificate;
+        GTlsCertificateFlags  tls_peer_certificate_errors;
 };
 
 struct _SoupServerMessageClass {
         GObjectClass parent_class;
 };
 
-G_DEFINE_TYPE (SoupServerMessage, soup_server_message, G_TYPE_OBJECT)
+G_DEFINE_FINAL_TYPE (SoupServerMessage, soup_server_message, G_TYPE_OBJECT)
 
 enum {
         WROTE_INFORMATIONAL,
@@ -101,6 +97,17 @@ enum {
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
+enum {
+	PROP_0,
+
+	PROP_TLS_PEER_CERTIFICATE,
+	PROP_TLS_PEER_CERTIFICATE_ERRORS,
+
+	LAST_PROPERTY
+};
+
+static GParamSpec *properties[LAST_PROPERTY] = { NULL, };
+
 static void
 soup_server_message_init (SoupServerMessage *msg)
 {
@@ -141,12 +148,32 @@ soup_server_message_finalize (GObject *object)
         G_OBJECT_CLASS (soup_server_message_parent_class)->finalize (object);
 }
 
+static void
+soup_server_message_get_property (GObject *object, guint prop_id,
+                                  GValue *value, GParamSpec *pspec)
+{
+	SoupServerMessage *msg = SOUP_SERVER_MESSAGE (object);
+
+	switch (prop_id) {
+	case PROP_TLS_PEER_CERTIFICATE:
+		g_value_set_object (value, msg->tls_peer_certificate);
+		break;
+	case PROP_TLS_PEER_CERTIFICATE_ERRORS:
+		g_value_set_flags (value, msg->tls_peer_certificate_errors);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
 static void
 soup_server_message_class_init (SoupServerMessageClass *klass)
 {
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
         object_class->finalize = soup_server_message_finalize;
+        object_class->get_property = soup_server_message_get_property;
 
         /**
          * SoupServerMessage::wrote-informational:
@@ -186,11 +213,11 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
          * Emitted immediately after writing a body chunk for a message.
          *
          * Note that this signal is not parallel to
-         * #SoupServerMessage::got-chunk; it is emitted only when a complete
-         * chunk (added with soup_message_body_append() or
-         * soup_message_body_append_bytes()) has been written. To get
+         * [signal@ServerMessage::got-chunk]; it is emitted only when a complete
+         * chunk (added with [method@MessageBody.append] or
+         * [method@MessageBody.append_bytes] has been written. To get
          * more useful continuous progress information, use
-         * #SoupServerMessage::wrote-body-data.
+         * [signal@ServerMessage::wrote-body-data].
          */
         signals[WROTE_CHUNK] =
                 g_signal_new ("wrote-chunk",
@@ -255,10 +282,10 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
          * @msg: the message
          * @chunk: the just-read chunk
          *
-         * Emitted after receiving a chunk of a message body. Note
-         * that "chunk" in this context means any subpiece of the
-         * body, not necessarily the specific HTTP 1.1 chunks sent by
-         * the other side.
+         * Emitted after receiving a chunk of a message body.
+         *
+         * Note that "chunk" in this context means any subpiece of the body, not
+         * necessarily the specific HTTP 1.1 chunks sent by the other side.
          */
         signals[GOT_CHUNK] =
                 g_signal_new ("got-chunk",
@@ -290,7 +317,7 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
          * @msg: the message
          *
          * Emitted when all HTTP processing is finished for a message.
-         * (After #SoupServerMessage::wrote-body).
+         * (After [signal@ServerMessage::wrote-body]).
          */
         signals[FINISHED] =
                 g_signal_new ("finished",
@@ -328,8 +355,8 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
 	 * @tls_errors.
 	 *
 	 * Returns: %TRUE to accept the TLS certificate and stop other
-	 *     handlers from being invoked, or %FALSE to propagate the
-	 *     event further.
+	 *   handlers from being invoked, or %FALSE to propagate the
+	 *   event further.
 	 */
 	signals[ACCEPT_CERTIFICATE] =
 		g_signal_new ("accept-certificate",
@@ -341,6 +368,38 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
 			      G_TYPE_BOOLEAN, 2,
 			      G_TYPE_TLS_CERTIFICATE,
 			      G_TYPE_TLS_CERTIFICATE_FLAGS);
+
+	/**
+	 * SoupServerMessage:tls-peer-certificate:
+	 *
+	 * The peer's #GTlsCertificate associated with the message
+	 *
+	 * Since: 3.2
+	 */
+        properties[PROP_TLS_PEER_CERTIFICATE] =
+		g_param_spec_object ("tls-peer-certificate",
+				     "TLS Peer Certificate",
+				     "The TLS peer certificate associated with the message",
+				     G_TYPE_TLS_CERTIFICATE,
+				     G_PARAM_READABLE |
+				     G_PARAM_STATIC_STRINGS);
+
+	/**
+	 * SoupServerMessage:tls-peer-certificate-errors:
+	 *
+	 * The verification errors on #SoupServerMessage:tls-peer-certificate
+	 *
+	 * Since: 3.2
+	 */
+        properties[PROP_TLS_PEER_CERTIFICATE_ERRORS] =
+		g_param_spec_flags ("tls-peer-certificate-errors",
+				    "TLS Peer Certificate Errors",
+				    "The verification errors on the message's TLS peer certificate",
+				    G_TYPE_TLS_CERTIFICATE_FLAGS, 0,
+				    G_PARAM_READABLE |
+				    G_PARAM_STATIC_STRINGS);
+
+        g_object_class_install_properties (object_class, LAST_PROPERTY, properties);
 }
 
 static void
@@ -361,6 +420,31 @@ socket_accept_certificate (SoupServerMessage    *msg,
 	return accept;
 }
 
+static void
+soup_server_message_set_tls_peer_certificate (SoupServerMessage   *msg,
+                                              GTlsCertificate     *tls_certificate,
+                                              GTlsCertificateFlags tls_errors)
+{
+        if (msg->tls_peer_certificate == tls_certificate && msg->tls_peer_certificate_errors == tls_errors)
+                return;
+
+        g_clear_object (&msg->tls_peer_certificate);
+        msg->tls_peer_certificate = tls_certificate ? g_object_ref (tls_certificate) : NULL;
+        msg->tls_peer_certificate_errors = tls_errors;
+        g_object_notify_by_pspec (G_OBJECT (msg), properties[PROP_TLS_PEER_CERTIFICATE]);
+        g_object_notify_by_pspec (G_OBJECT (msg), properties[PROP_TLS_PEER_CERTIFICATE_ERRORS]);
+}
+
+static void
+re_emit_tls_certificate_changed (SoupServerMessage *msg,
+                                 GParamSpec        *pspec,
+                                 SoupSocket        *sock)
+{
+        soup_server_message_set_tls_peer_certificate (msg,
+                                                      soup_socket_get_tls_certificate (sock),
+                                                      soup_socket_get_tls_certificate_errors (sock));
+}
+
 SoupServerMessage *
 soup_server_message_new (SoupSocket *sock)
 {
@@ -378,6 +462,9 @@ soup_server_message_new (SoupSocket *sock)
         g_signal_connect_object (sock, "accept-certificate",
                                  G_CALLBACK (socket_accept_certificate),
                                  msg, G_CONNECT_SWAPPED);
+        g_signal_connect_object (sock, "notify::tls-certificate",
+                                 G_CALLBACK (re_emit_tls_certificate_changed),
+                                 msg, G_CONNECT_SWAPPED);
 
         return msg;
 }
@@ -680,9 +767,9 @@ soup_server_message_set_http_version (SoupServerMessage *msg,
  * soup_server_message_get_reason_phrase:
  * @msg: a #SoupServerMessage:
  *
- * Get the HTTP reason phrase of @msg or %NULL.
+ * Get the HTTP reason phrase of @msg.
  *
- * Returns: the reason phrase.
+ * Returns: (nullable): the reason phrase.
  */
 const char *
 soup_server_message_get_reason_phrase (SoupServerMessage *msg)
@@ -714,9 +801,10 @@ soup_server_message_get_status (SoupServerMessage *msg)
  * @status_code: an HTTP status code
  * @reason_phrase: (nullable): a reason phrase
  *
- * Sets @msg's status code to @status_code. If @status_code is a
- * known value and @reason_phrase is %NULL, the reason_phrase will
- * be set automatically.
+ * Sets @msg's status code to @status_code.
+ *
+ * If @status_code is a known value and @reason_phrase is %NULL, the
+ * reason_phrase will be set automatically.
  **/
 void
 soup_server_message_set_status (SoupServerMessage *msg,
@@ -751,9 +839,9 @@ soup_server_message_get_uri (SoupServerMessage *msg)
 /**
  * soup_server_message_set_response:
  * @msg: the message
- * @content_type: (allow-none): MIME Content-Type of the body
+ * @content_type: (nullable): MIME Content-Type of the body
  * @resp_use: a #SoupMemoryUse describing how to handle @resp_body
- * @resp_body: (allow-none) (array length=resp_length) (element-type guint8):
+ * @resp_body: (nullable) (array length=resp_length) (element-type guint8):
  *   a data buffer containing the body of the message response.
  * @resp_length: the byte length of @resp_body.
  *
@@ -792,13 +880,13 @@ soup_server_message_set_response (SoupServerMessage *msg,
  * @redirect_uri: the URI to redirect @msg to
  *
  * Sets @msg's status_code to @status_code and adds a Location header
- * pointing to @redirect_uri. Use this from a #SoupServer when you
+ * pointing to @redirect_uri. Use this from a [class@Server] when you
  * want to redirect the client to another URI.
  *
  * @redirect_uri can be a relative URI, in which case it is
  * interpreted relative to @msg's current URI. In particular, if
  * @redirect_uri is just a path, it will replace the path
- * <emphasis>and query</emphasis> of @msg's URI.
+ * *and query* of @msg's URI.
  */
 void
 soup_server_message_set_redirect (SoupServerMessage *msg,
@@ -825,7 +913,7 @@ soup_server_message_set_redirect (SoupServerMessage *msg,
  * soup_server_message_get_socket:
  * @msg: a #SoupServerMessage
  *
- * Retrieves the #GSocket that @msg is associated with.
+ * Retrieves the [class@Gio.Socket] that @msg is associated with.
  *
  * If you are using this method to observe when multiple requests are
  * made on the same persistent HTTP connection (eg, as the ntlm-test
@@ -835,7 +923,7 @@ soup_server_message_set_redirect (SoupServerMessage *msg,
  * previously-destroyed socket to represent a new socket.
  *
  * Returns: (nullable) (transfer none): the #GSocket that @msg is
- * associated with, %NULL if you used soup_server_accept_iostream().
+ *   associated with, %NULL if you used [method@Server.accept_iostream].
  */
 GSocket *
 soup_server_message_get_socket (SoupServerMessage *msg)
@@ -849,12 +937,12 @@ soup_server_message_get_socket (SoupServerMessage *msg)
  * soup_server_message_get_remote_address:
  * @msg: a #SoupServerMessage
  *
- * Retrieves the #GSocketAddress associated with the remote end
+ * Retrieves the [class@Gio.SocketAddress] associated with the remote end
  * of a connection.
  *
  * Returns: (nullable) (transfer none): the #GSocketAddress
- *     associated with the remote end of a connection, it may be
- *     %NULL if you used soup_server_accept_iostream().
+ *   associated with the remote end of a connection, it may be
+ *   %NULL if you used [class@Server.accept_iostream].
  */
 GSocketAddress *
 soup_server_message_get_remote_address (SoupServerMessage *msg)
@@ -875,12 +963,12 @@ soup_server_message_get_remote_address (SoupServerMessage *msg)
  * soup_server_message_get_local_address:
  * @msg: a #SoupServerMessage
  *
- * Retrieves the #GSocketAddress associated with the local end
+ * Retrieves the [class@Gio.SocketAddress] associated with the local end
  * of a connection.
  *
  * Returns: (nullable) (transfer none): the #GSocketAddress
- *     associated with the local end of a connection, it may be
- *     %NULL if you used soup_server_accept_iostream().
+ *   associated with the local end of a connection, it may be
+ *   %NULL if you used [method@Server.accept_iostream].
  */
 GSocketAddress *
 soup_server_message_get_local_address (SoupServerMessage *msg)
@@ -905,8 +993,8 @@ soup_server_message_get_local_address (SoupServerMessage *msg)
  * connection.
  *
  * Returns: (nullable): the IP address associated with the remote
- *     end of a connection, it may be %NULL if you used
- *     soup_server_accept_iostream().
+ *   end of a connection, it may be %NULL if you used
+ *   [method@Server.accept_iostream].
  */
 const char *
 soup_server_message_get_remote_host (SoupServerMessage *msg)
@@ -937,12 +1025,13 @@ soup_server_message_get_remote_host (SoupServerMessage *msg)
  * soup_server_message_steal_connection:
  * @msg: a #SoupServerMessage
  *
- * "Steals" the HTTP connection associated with @msg from its
- * #SoupServer. This happens immediately, regardless of the current
- * state of the connection; if the response to @msg has not yet finished
- * being sent, then it will be discarded; you can steal the connection from a
- * #SoupServerMessage::wrote-informational or #SoupServerMessage::wrote-body signal
- * handler if you need to wait for part or all of the response to be sent.
+ * "Steals" the HTTP connection associated with @msg from its #SoupServer. This
+ * happens immediately, regardless of the current state of the connection; if
+ * the response to @msg has not yet finished being sent, then it will be
+ * discarded; you can steal the connection from a
+ * [signal@ServerMessage::wrote-informational] or
+ * [signal@ServerMessage::wrote-body] signal handler if you need to wait for
+ * part or all of the response to be sent.
  *
  * Note that when calling this function from C, @msg will most
  * likely be freed as a side effect.
@@ -972,3 +1061,44 @@ soup_server_message_steal_connection (SoupServerMessage *msg)
 
         return stream;
 }
+
+/**
+ * soup_server_message_get_tls_peer_certificate:
+ * @msg: a #SoupMessage
+ *
+ * Gets the peer's #GTlsCertificate associated with @msg's connection.
+ * Note that this is not set yet during the emission of
+ * SoupServerMessage::accept-certificate signal.
+ *
+ * Returns: (transfer none) (nullable): @msg's TLS peer certificate,
+ *    or %NULL if @msg's connection is not SSL.
+ *
+ * Since: 3.2
+ */
+GTlsCertificate *
+soup_server_message_get_tls_peer_certificate (SoupServerMessage *msg)
+{
+        g_return_val_if_fail (SOUP_IS_SERVER_MESSAGE (msg), NULL);
+
+        return msg->tls_peer_certificate;
+}
+
+/**
+ * soup_server_message_get_tls_peer_certificate_errors:
+ * @msg: a #SoupMessage
+ *
+ * Gets the errors associated with validating @msg's TLS peer certificate.
+ * Note that this is not set yet during the emission of
+ * SoupServerMessage::accept-certificate signal.
+ *
+ * Returns: a #GTlsCertificateFlags with @msg's TLS peer certificate errors.
+ *
+ * Since: 3.2
+ */
+GTlsCertificateFlags
+soup_server_message_get_tls_peer_certificate_errors (SoupServerMessage *msg)
+{
+        g_return_val_if_fail (SOUP_IS_SERVER_MESSAGE (msg), 0);
+
+        return msg->tls_peer_certificate_errors;
+}
