@@ -47,6 +47,7 @@ typedef struct {
         GTlsCertificate *tls_client_cert;
 
 	GCancellable *cancellable;
+        GThread *owner;
 } SoupConnectionPrivate;
 
 G_DEFINE_FINAL_TYPE_WITH_PRIVATE (SoupConnection, soup_connection, G_TYPE_OBJECT)
@@ -76,13 +77,14 @@ enum {
         PROP_TLS_PROTOCOL_VERSION,
         PROP_TLS_CIPHERSUITE_NAME,
         PROP_FORCE_HTTP_VERSION,
+        PROP_CONTEXT,
 
 	LAST_PROPERTY
 };
 
 static GParamSpec *properties[LAST_PROPERTY] = { NULL, };
 
-static void stop_idle_timer (SoupConnectionPrivate *priv);
+static gboolean idle_timeout (gpointer conn);
 
 /* Number of seconds after which we close a connection that hasn't yet
  * been used.
@@ -96,6 +98,7 @@ soup_connection_init (SoupConnection *conn)
 
         priv->http_version = SOUP_HTTP_1_1;
         priv->force_http_version = G_MAXUINT8;
+        priv->owner = g_thread_self ();
 }
 
 static void
@@ -133,7 +136,11 @@ soup_connection_dispose (GObject *object)
 	SoupConnection *conn = SOUP_CONNECTION (object);
 	SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-	stop_idle_timer (priv);
+        if (priv->idle_timeout_src) {
+                g_source_destroy (priv->idle_timeout_src);
+                g_source_unref (priv->idle_timeout_src);
+                priv->idle_timeout_src = NULL;
+        }
 
 	G_OBJECT_CLASS (soup_connection_parent_class)->dispose (object);
 }
@@ -160,6 +167,13 @@ soup_connection_set_property (GObject *object, guint prop_id,
 	case PROP_FORCE_HTTP_VERSION:
 		priv->force_http_version = g_value_get_uchar (value);
 		break;
+        case PROP_CONTEXT:
+                priv->idle_timeout_src = g_timeout_source_new (0);
+                g_source_set_ready_time (priv->idle_timeout_src, -1);
+                g_source_set_name (priv->idle_timeout_src, "Soup connection idle timeout");
+                g_source_set_callback (priv->idle_timeout_src, idle_timeout, object, NULL);
+                g_source_attach (priv->idle_timeout_src, g_value_get_pointer (value));
+                break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -183,7 +197,7 @@ soup_connection_get_property (GObject *object, guint prop_id,
 		g_value_set_boxed (value, priv->socket_props);
 		break;
 	case PROP_STATE:
-		g_value_set_enum (value, priv->state);
+		g_value_set_enum (value, g_atomic_int_get (&priv->state));
 		break;
 	case PROP_SSL:
 		g_value_set_boolean (value, priv->ssl);
@@ -352,6 +366,12 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
                                     0, G_MAXUINT8, G_MAXUINT8,
                                     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
                                     G_PARAM_STATIC_STRINGS);
+        properties[PROP_CONTEXT] =
+                g_param_spec_pointer ("context",
+                                      "Context",
+                                      "The session main context",
+                                      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY |
+                                      G_PARAM_STATIC_STRINGS);
 
         g_object_class_install_properties (object_class, LAST_PROPERTY, properties);
 }
@@ -371,7 +391,7 @@ static gboolean
 idle_timeout (gpointer conn)
 {
 	soup_connection_disconnect (conn);
-	return FALSE;
+	return G_SOURCE_REMOVE;
 }
 
 static void
@@ -379,21 +399,14 @@ start_idle_timer (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-	if (priv->socket_props->idle_timeout > 0 && !priv->idle_timeout_src) {
-		priv->idle_timeout_src =
-			soup_add_timeout (g_main_context_get_thread_default (),
-					  priv->socket_props->idle_timeout * 1000,
-					  idle_timeout, conn);
-	}
-}
+	if (priv->socket_props->idle_timeout == 0)
+                return;
 
-static void
-stop_idle_timer (SoupConnectionPrivate *priv)
-{
-	if (priv->idle_timeout_src) {
-		g_source_destroy (priv->idle_timeout_src);
-                g_clear_pointer (&priv->idle_timeout_src, g_source_unref);
-	}
+        if (g_source_get_ready_time (priv->idle_timeout_src) >= 0)
+                return;
+
+        g_source_set_ready_time (priv->idle_timeout_src,
+                                 g_get_monotonic_time () + (guint64)priv->socket_props->idle_timeout * G_USEC_PER_SEC);
 }
 
 static void
@@ -402,11 +415,11 @@ soup_connection_set_state (SoupConnection     *conn,
 {
         SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-        if (priv->state == state)
+        if (g_atomic_int_get (&priv->state) == state)
                 return;
 
-        priv->state = state;
-        if (priv->state == SOUP_CONNECTION_IDLE)
+        g_atomic_int_set (&priv->state, state);
+        if (state == SOUP_CONNECTION_IDLE)
                 start_idle_timer (conn);
 
         g_object_notify_by_pspec (G_OBJECT (conn), properties[PROP_STATE]);
@@ -514,14 +527,13 @@ new_socket_client (SoupConnection *conn)
                                  G_CALLBACK (re_emit_socket_event),
                                  conn, 0);
 
-        if (!props->proxy_use_default && !props->proxy_resolver) {
-                g_socket_client_set_enable_proxy (client, FALSE);
-        } else {
-                if (props->proxy_resolver)
-                        g_socket_client_set_proxy_resolver (client, props->proxy_resolver);
-                g_socket_client_add_application_proxy (client, "http");
-        }
-
+	if (!props->proxy_use_default) {
+		if (props->proxy_resolver) {
+			g_socket_client_set_proxy_resolver (client, props->proxy_resolver);
+			g_socket_client_add_application_proxy (client, "http");
+		} else
+			g_socket_client_set_enable_proxy (client, FALSE);
+	}
         if (props->io_timeout)
                 g_socket_client_set_timeout (client, props->io_timeout);
         if (props->local_addr)
@@ -1004,7 +1016,8 @@ client_message_io_closed_cb (SoupConnection *conn,
  * soup_connection_disconnect:
  * @conn: a connection
  *
- * Disconnects @conn's socket and emits a %disconnected signal.
+ * Disconnects @conn's socket and emits a [signal@Socket::disconnected] signal.
+ *
  * After calling this, @conn will be essentially useless.
  **/
 void
@@ -1012,7 +1025,7 @@ soup_connection_disconnect (SoupConnection *conn)
 {
         SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-        if (priv->state == SOUP_CONNECTION_DISCONNECTED)
+        if (g_atomic_int_get (&priv->state) == SOUP_CONNECTION_DISCONNECTED)
                 return;
 
         soup_connection_set_state (conn, SOUP_CONNECTION_DISCONNECTED);
@@ -1107,7 +1120,8 @@ soup_connection_is_idle_open (SoupConnection *conn)
 {
         SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-        g_assert (priv->state == SOUP_CONNECTION_IDLE);
+        if (g_atomic_int_get (&priv->state) != SOUP_CONNECTION_IDLE)
+                return FALSE;
 
 	if (!g_socket_is_connected (soup_connection_get_socket (conn)))
 		return FALSE;
@@ -1123,7 +1137,7 @@ soup_connection_get_state (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-	return priv->state;
+	return g_atomic_int_get (&priv->state);
 }
 
 void
@@ -1132,25 +1146,29 @@ soup_connection_set_in_use (SoupConnection *conn,
 {
         SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-        g_assert (in_use || priv->in_use > 0);
+        g_assert (in_use || g_atomic_int_get (&priv->in_use) > 0);
 
-        if (in_use)
-                priv->in_use++;
-        else
-                priv->in_use--;
+        if (in_use) {
+                g_atomic_int_inc (&priv->in_use);
+                if (g_atomic_int_compare_and_exchange (&priv->state, SOUP_CONNECTION_IDLE, SOUP_CONNECTION_IN_USE)) {
+                        priv->owner = g_thread_self ();
+                        soup_client_message_io_owner_changed (priv->io_data);
+                        g_object_notify_by_pspec (G_OBJECT (conn), properties[PROP_STATE]);
+                }
 
-        if (priv->in_use > 0) {
-                if (priv->state == SOUP_CONNECTION_IDLE)
-                        soup_connection_set_state (conn, SOUP_CONNECTION_IN_USE);
                 return;
         }
 
-        clear_proxy_msg (conn);
+        g_assert (g_atomic_int_get (&priv->state) != SOUP_CONNECTION_IDLE);
 
-        if (soup_connection_is_reusable (conn))
-                soup_connection_set_state (conn, SOUP_CONNECTION_IDLE);
-        else
-                soup_connection_disconnect (conn);
+        if (g_atomic_int_dec_and_test (&priv->in_use)) {
+                clear_proxy_msg (conn);
+
+                if (soup_connection_is_reusable (conn))
+                        soup_connection_set_state (conn, SOUP_CONNECTION_IDLE);
+                else
+                        soup_connection_disconnect (conn);
+        }
 }
 
 SoupClientMessageIO *
@@ -1159,10 +1177,10 @@ soup_connection_setup_message_io (SoupConnection *conn,
 {
         SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
 
-        g_assert (priv->state == SOUP_CONNECTION_IN_USE);
+        g_assert (g_atomic_int_get (&priv->state) == SOUP_CONNECTION_IN_USE);
 
         priv->unused_timeout = 0;
-        stop_idle_timer (priv);
+        g_source_set_ready_time (priv->idle_timeout_src, -1);
 
         if (priv->proxy_uri && soup_message_get_method (msg) == SOUP_METHOD_CONNECT)
                 set_proxy_msg (conn, msg);
@@ -1350,3 +1368,11 @@ soup_connection_is_reusable (SoupConnection *conn)
 
         return priv->io_data && soup_client_message_io_is_reusable (priv->io_data);
 }
+
+GThread *
+soup_connection_get_owner (SoupConnection *conn)
+{
+        SoupConnectionPrivate *priv = soup_connection_get_instance_private (conn);
+
+        return priv->owner;
+}
