@@ -6,6 +6,7 @@
 #include "test-utils.h"
 #include "soup-connection.h"
 #include "soup-session-private.h"
+#include "soup-message-headers-private.h"
 
 SoupServer *server;
 GUri *base_uri;
@@ -55,6 +56,15 @@ server_callback (SoupServer        *server,
                 g_source_unref (timeout);
 	}
 
+        if (!strcmp (path, "/invalid_utf8_headers")) {
+                SoupMessageHeaders *headers = soup_server_message_get_response_headers (msg);
+                const char *invalid_utf8_value = "\xe2\x82\xa0gh\xe2\xffjl";
+
+                /* Purposefully insert invalid utf8 data */
+                g_assert_false (g_utf8_validate (invalid_utf8_value, -1, NULL));
+                soup_message_headers_append (headers, "InvalidValue", invalid_utf8_value);
+        }
+
 	soup_server_message_set_status (msg, SOUP_STATUS_OK, NULL);
 	if (!strcmp (g_uri_get_host (uri), "foo")) {
 		soup_server_message_set_response (msg, "text/plain",
@@ -807,14 +817,14 @@ do_remote_address_test (void)
         g_bytes_unref (body);
         g_object_unref (msg2);
 
-        /* We get a new one if we force a new connection */
+        /* We get the same one if we force a new connection */
         msg2 = soup_message_new_from_uri (SOUP_METHOD_GET, base_uri);
         soup_message_add_flags (msg2, SOUP_MESSAGE_NEW_CONNECTION);
         g_assert_null (soup_message_get_remote_address (msg2));
         body = soup_test_session_async_send (session, msg2, NULL, NULL);
         g_assert_nonnull (soup_message_get_remote_address (msg2));
         g_assert_cmpuint (soup_message_get_connection_id (msg1), !=, soup_message_get_connection_id (msg2));
-        g_assert_false (soup_message_get_remote_address (msg1) == soup_message_get_remote_address (msg2));
+        g_assert_true (soup_message_get_remote_address (msg1) == soup_message_get_remote_address (msg2));
         g_bytes_unref (body);
         g_object_unref (msg2);
 
@@ -861,6 +871,106 @@ do_new_request_on_redirect_test (void)
         soup_test_session_abort_unref (session);
 }
 
+static void
+wrote_informational_check_content_length (SoupServerMessage *msg,
+                                          gpointer           user_data)
+{
+        SoupMessageHeaders *response_headers;
+
+        response_headers = soup_server_message_get_response_headers (msg);
+        g_assert_null (soup_message_headers_get_one (response_headers, "Content-Length"));
+}
+
+static void
+upgrade_server_check_content_length_callback (SoupServer        *server,
+                                              SoupServerMessage *msg,
+                                              const char        *path,
+                                              GHashTable        *query,
+                                              gpointer           data)
+{
+        SoupMessageHeaders *request_headers;
+
+        if (soup_server_message_get_method (msg) != SOUP_METHOD_GET) {
+                soup_server_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED, NULL);
+                return;
+        }
+
+        soup_server_message_set_status (msg, SOUP_STATUS_SWITCHING_PROTOCOLS, NULL);
+
+        request_headers = soup_server_message_get_request_headers (msg);
+        soup_message_headers_append (request_headers, "Upgrade", "ECHO");
+        soup_message_headers_append (request_headers, "Connection", "upgrade");
+
+        g_signal_connect (msg, "wrote-informational",
+                          G_CALLBACK (wrote_informational_check_content_length), NULL);
+}
+
+static void
+switching_protocols_check_length (SoupMessage *msg,
+                                  gpointer     user_data)
+{
+        SoupMessageHeaders *response_headers;
+
+        response_headers = soup_message_get_response_headers (msg);
+        g_assert_null (soup_message_headers_get_one (response_headers, "Content-Length"));
+}
+
+static void
+do_response_informational_content_length_test (void)
+{
+        SoupServer *server;
+        SoupSession *session;
+        SoupMessage *msg;
+        SoupMessageHeaders *request_headers;
+
+        server = soup_test_server_new (SOUP_TEST_SERVER_IN_THREAD);
+        soup_server_add_handler (server, NULL, upgrade_server_check_content_length_callback, NULL, NULL);
+
+        session = soup_test_session_new (NULL);
+        msg = soup_message_new_from_uri ("GET", base_uri);
+        request_headers = soup_message_get_request_headers (msg);
+        soup_message_headers_append (request_headers, "Upgrade", "echo");
+        soup_message_headers_append (request_headers, "Connection", "upgrade");
+
+        soup_message_add_status_code_handler (msg, "got-informational",
+                                              SOUP_STATUS_SWITCHING_PROTOCOLS,
+                                              G_CALLBACK (switching_protocols_check_length), NULL);
+
+        soup_test_session_send_message (session, msg);
+        g_object_unref (msg);
+
+        soup_test_session_abort_unref (session);
+        soup_test_server_quit_unref (server);
+}
+
+static void
+do_invalid_utf8_headers_test (void)
+{
+        SoupSession *session;
+        SoupMessage *msg;
+        GUri *uri;
+        SoupMessageHeaders *headers;
+        guint status;
+        const char *header_value;
+
+        session = soup_test_session_new (NULL);
+
+        uri = g_uri_parse_relative (base_uri, "/invalid_utf8_headers", SOUP_HTTP_URI_FLAGS, NULL);
+        msg = soup_message_new_from_uri ("GET", uri);
+
+        status = soup_test_session_send_message (session, msg);
+        g_assert_cmpuint (status, ==, SOUP_STATUS_OK);
+
+        headers = soup_message_get_response_headers (msg);
+        header_value = soup_message_headers_get_one (headers, "InvalidValue");
+        g_assert_nonnull (header_value);
+        g_assert_true (g_utf8_validate (header_value, -1, NULL));
+
+        g_object_unref (msg);
+        g_uri_unref (uri);
+        soup_test_session_abort_unref (session);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -896,6 +1006,8 @@ main (int argc, char **argv)
         g_test_add_func ("/misc/connection-id", do_connection_id_test);
         g_test_add_func ("/misc/remote-address", do_remote_address_test);
         g_test_add_func ("/misc/new-request-on-redirect", do_new_request_on_redirect_test);
+        g_test_add_func ("/misc/response/informational/content-length", do_response_informational_content_length_test);
+        g_test_add_func ("/misc/invalid-utf8-headers", do_invalid_utf8_headers_test);
 
 	ret = g_test_run ();
 
